# Path finder demo
# http://www.willmcgugan.com


SCREEN_SIZE = (768, 576)
GRID_SQUARE_SIZE = (64, 64)
background_image_filename = 'sea.jpg'
block_image_filename = 'bigblock.png'
mouse_image_filename = "mousecursor.png"
fish_image_filename = "smallfugu.png"

import pygame
from pygame.locals import *
from copy import copy
from math import sqrt
from gameobjects.vector2 import Vector2


class Square(object):
    
    def __init__(self, coord):
        
        self.coord = coord
        self.blocked = False
        self.parent = None        
        
    def reset(self):
        self.parent = None        


class Grid(object):
    
    def __init__(self, width, height):
        
        self.width = width
        self.height = height
        
        self.rows = []
        for row in xrange(height):            
            row_squares = []            
            for col in xrange(width):                
                row_squares.append( Square((col, row)) )
            self.rows.append(row_squares)                
        
    def get(self, coord):
        
        if not self.is_valid_coordinate(coord):
            return None
        x, y = coord
        return self.rows[y][x]
    
    def set(self, coord, blocked):
        
        if self.is_valid_coordinate(coord):
            x, y = coord
            self.rows[y][x] = blocked
    
    def is_valid_coordinate(self, coord):
        
        x, y = coord
        if x < 0 or y < 0 or x >= self.width or y >= self.height:
            return False
        return True

    def reset(self):
        
        for row in self.rows:
            for square in row:
                square.reset()
    
        
    def find_route(self, start_coord, dest_coord):
        
        self.reset()
        
        visited = set()
        open_squares = []
        
        open_squares.append(start_coord)    
        
        steps = ( (0,  +1), (+1,  0), (0,  -1), (-1,  0 ),
                  (+1, -1), (+1, +1), (-1, +1), (-1, -1) )
        
        while open_squares:
            
            coord = open_squares.pop(0)        
            
            if coord == dest_coord:
                path = []            
                while coord != start_coord:
                    path.append(coord)
                    square = self.get(coord)                
                    parent_square = self.get(square.parent)
                    coord = square.parent                
                return path[::-1]
     
            for step in steps:
                
                new_coord = (coord[0]+step[0], coord[1]+step[1])
                
                if new_coord in visited:
                    continue
                
                visited.add(new_coord)            
                square = self.get(new_coord)
                
                if square is None or square.blocked:
                    continue
                
                self.get(new_coord).parent = coord
                open_squares.append(new_coord)
                
        return None
    

    def render(self, surface, block_surface):
        
        for y in xrange(self.height):
            for x in xrange(self.width):            
                coord = (x, y)
                if self.get(coord).blocked:
                    render_x = x * GRID_SQUARE_SIZE[0]
                    render_y = y * GRID_SQUARE_SIZE[1]                
                    surface.blit(block_surface, (render_x, render_y))
                
   
def run():
    
    instructions_txt = "Draw map with left mouse button. "
    instructions_txt += "Press right mouse button (or Shift+LMB) to swim."
    title = "Path Finding Demo"
            
    fish_speed = 300.
    
    pygame.init()
    screen = pygame.display.set_mode(SCREEN_SIZE, 0, 32)
    pygame.display.set_caption(title)
    
    default_font = pygame.font.get_default_font()
    font = pygame.font.SysFont(default_font, 26)    
    instructions = font.render(instructions_txt, True, (255, 255, 255))
    
    pygame.mouse.set_visible(False)
    
    background = pygame.image.load(background_image_filename).convert()
    block = pygame.image.load(block_image_filename).convert_alpha()
    mouse = pygame.image.load(mouse_image_filename).convert_alpha()
    fish = pygame.image.load(fish_image_filename).convert_alpha()
        
    width, height = SCREEN_SIZE
    grid = Grid(width/64, height/64)
    
    def pos_to_coord(pos):        
        coord_x = event.pos[0] / GRID_SQUARE_SIZE[0]
        coord_y = event.pos[1] / GRID_SQUARE_SIZE[1]        
        return (coord_x, coord_y)
    
    def coord_to_pos(coord):
        pos_x = coord[0] * GRID_SQUARE_SIZE[0]
        pos_y = coord[1] * GRID_SQUARE_SIZE[1]
        return (pos_x, pos_y)
        
    fish_coord = (0, 0)
    fish_pos = Vector2(*coord_to_pos(fish_coord))
    route = []
    clock = pygame.time.Clock()
        
    while True:
        
        for event in pygame.event.get():
            
            if event.type == QUIT:
                return
            
            if event.type == MOUSEBUTTONDOWN and not route:
                
                mouse_coord = pos_to_coord(event.pos)
                clicked_square = grid.get(mouse_coord)
                
                if mouse_coord == fish_coord or clicked_square is None:
                    continue
                                
                shift_pressed = pygame.key.get_mods() & (KMOD_SHIFT)
                
                if event.button == 3 or (event.button == 1 and shift_pressed):
                    
                    if not clicked_square.blocked:                        
                        route = grid.find_route(fish_coord, mouse_coord)                        
                        if route is not None:                            
                            pygame.display.set_caption("Swimming to destination!")
                            fish_coord = mouse_coord
                            
                elif event.button == 1:                                                            
                    clicked_square.blocked = not clicked_square.blocked
            
        screen.blit(background, (0,0))        
        
        grid.render(screen, block)
        
        time_passed = clock.tick()
        time_passed_seconds = time_passed / 1000.0
 
        if route:
            
            for coord in route:
                x, y = coord_to_pos(coord)
                x += GRID_SQUARE_SIZE[0] / 2
                y += GRID_SQUARE_SIZE[1] / 2
                pygame.draw.circle(screen, (0, 255, 0), (x, y), 6)
            
            destination = Vector2( *coord_to_pos(route[0]) )
            heading = Vector2.from_points(fish_pos, destination)
            distance = heading.get_length()
            heading.normalize()       
            
            if fish_speed * time_passed_seconds > distance:
                movement = heading * distance 
                route = route[1:]
                if not route:
                    pygame.display.set_caption(title)
            else:
                movement = heading * fish_speed * time_passed_seconds
                
            fish_pos += movement
                
        screen.blit(fish, fish_pos)
        
        mouse_pos = pygame.mouse.get_pos()
        screen.blit(mouse, mouse_pos)
        
        screen.blit(instructions, (5, 5))
        
        pygame.display.update()
        
        

if __name__ == "__main__":
    run()